from flo_ai.models.flo_team import FloTeamBuilder
from flo_ai.models.flo_agent import FloAgentBuilder, FloAgent
from flo_ai.models.flo_supervisor import FloSupervisorBuilder
from flo_ai.yaml.flo_team_builder import (FloRoutedTeamConfig, TeamConfig,
                                        AgentConfig, FloAgentConfig)
from flo_ai.models.flo_executable import ExecutableFlo
from flo_ai.models.flo_planner import FloPlannerBuilder
from flo_ai.state.flo_session import FloSession
from typing import Union, List

def build_supervised_team(
        session: FloSession,
        flo_config: Union[FloRoutedTeamConfig, FloAgentConfig]) -> ExecutableFlo:
    if isinstance(flo_config, FloRoutedTeamConfig):
        team_config: TeamConfig = flo_config.team
        team = parse_and_build_subteams(session, team_config, session.tools)    
        return team
    elif isinstance(flo_config, FloAgentConfig):
        agent_config: AgentConfig = flo_config.agent
        agent = create_agent(session, agent_config, session.tools)
        return agent

def parse_and_build_subteams(
        session: FloSession,
        team: TeamConfig,
        tool_map) -> ExecutableFlo:
    flo_team = None
    if team.agents:
        agents = []
        for agent in team.agents:
            flo_agent: FloAgent = create_agent(session, agent, tool_map)
            agents.append(flo_agent)
        router = create_router(session, team, agents)
        flo_team = FloTeamBuilder(
            session=session,
            name=team.name,
            supervisor=router
        ).build()
        if team.planner is not None:
            return FloPlannerBuilder(session, team.planner.name, flo_team).build()
    else:
        flo_teams = []
        for subteam in team.subteams:
            flo_subteam = parse_and_build_subteams(session, subteam, tool_map)
            flo_teams.append(flo_subteam)
        router = create_router(session, team, agents)
        flo_team = FloTeamBuilder(
            session=session,
            name=team.name,
            supervisor=router
        ).build()
    return flo_team

def create_router(session: FloSession, team_config: TeamConfig, agents: List[FloAgent]):
    if (team_config.router.kind == "supervisor"):
            return FloSupervisorBuilder(session, team_config.router.name, agents).build()
    else:
        raise Exception("Unknown router type")
        
def create_agent(session: FloSession, agent: AgentConfig, tool_map) -> FloAgent:
    tools = [tool_map[tool.name] for tool in agent.tools]
    flo_agent: FloAgent = FloAgentBuilder(
        session,
        agent.name, 
        agent.job,
        tools
    ).build()
    return flo_agent




from typing import List, Union

from langchain.chains.llm import LLMChain
from langchain_core.vectorstores import VectorStoreRetriever
from langchain.output_parsers import PydanticOutputParser
from langchain_core.prompts import PromptTemplate
from pydantic import BaseModel, Field
from flo_ai.state.flo_session import FloSession
from langchain.retrievers.multi_query import MultiQueryRetriever

class LineList(BaseModel):
    lines: List[str] = Field(description="Lines of text")


class LineListOutputParser(PydanticOutputParser):
    def __init__(self) -> None:
        super().__init__(pydantic_object=LineList)

    def parse(self, text: str) -> LineList:
        lines = text.strip().split("\n")
        return LineList(lines=lines)
    
class FloMultiQueryRetriever():
    def __init__(self, retriever) -> None:
        self.retriever = retriever

class FloMultiQueryRetriverBuilder():

    def __init__(self, 
                 session: FloSession, 
                 retriver: VectorStoreRetriever,
                 query_prompt: Union[str, None] = None) -> None:
        self.session = session
        self.retriver = retriver
        self.output_parser = LineListOutputParser()

        self.prompt = PromptTemplate(
            input_variables=["question"],
            template="""You are an AI language model assistant. Your task is to generate three 
            different versions of the given user question to retrieve relevant documents from a vector 
            database. By generating multiple perspectives on the user question, your goal is to help
            the user overcome some of the limitations of the distance-based similarity search. 
            Provide these alternative questions separated by newlines.
            Original question: {question}""" if query_prompt is None else query_prompt,
        )

    def build(self):
        multi_query_retriever = MultiQueryRetriever.from_llm(
            retriever=self.retriver, 
            llm=self.session.llm,
            prompt=self.prompt
        )
        return FloMultiQueryRetriever(multi_query_retriever)




from typing import Optional
from langchain_core.language_models import BaseLanguageModel
from langchain.tools import Tool
from langgraph.graph import END, StateGraph
from langgraph.graph.graph import CompiledGraph
from langgraph.prebuilt import ToolNode
from flo_ai.state.flo_state import TeamFloAgentState
from langgraph.prebuilt import tools_condition
from flo_ai.models.flo_executable import ExecutableFlo
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain import hub
from flo_ai.helpers.utils import randomize_name

class FloRag(ExecutableFlo):
    def __init__(self, 
                 name: str, 
                 graph: CompiledGraph) -> None:
        super().__init__(name, "team")
        self.graph = graph

    def draw(self, xray=True):
        return self.graph.get_graph(xray=xray).draw_mermaid_png()

class FloRagBuilder:

    def __init__(self, 
                 name: str, 
                 tools: list[Tool], 
                 llm: BaseLanguageModel,
                 prompt: Optional[ChatPromptTemplate] = None) -> None:
        self.name = randomize_name(name)
        self.llm = llm
        self.tools = tools
        self.prompt = hub.pull("rlm/rag-prompt") if prompt is None else prompt
    
    def retriever_agent(self, state: TeamFloAgentState):
        messages = state["messages"]
        model = self.llm.bind_tools(self.tools)
        response = model.invoke(messages)
        # We return a list, because this will get added to the existing list
        return { "messages": [response] }
    
    def generate(self, state: TeamFloAgentState):
        messages = state["messages"]
        question = messages[0].content
        last_message = messages[-1]

        question = messages[0].content
        docs = last_message.content

        # Chain
        rag_chain = self.prompt | self.llm

        # Run
        response = rag_chain.invoke({"context": docs, "question": question})

        return {"messages": [response] }
    
    def build(self) -> FloRag:
        retrieve = ToolNode(self.tools)

        workflow = StateGraph(TeamFloAgentState)
        workflow.add_node("agent", self.retriever_agent)
        workflow.add_node("retrieve", retrieve)  # retrieval
        workflow.add_node("generate", self.generate)
        workflow.add_conditional_edges(
            "agent",
            tools_condition,
            {
                "tools": "retrieve",
                END: END,
            },
        )

        workflow.add_edge("retrieve", "generate")
        workflow.add_edge("generate", END)

        workflow.set_entry_point("agent")
        graph = workflow.compile()
        return FloRag(self.name, graph=graph)



from typing import Annotated, List, Sequence, TypedDict
from langchain_core.messages import BaseMessage
from typing import List, Tuple, Annotated, TypedDict, Dict

import operator

# The agent state is the input to each node in the graph
class TeamFloAgentState(TypedDict):
    # The annotation tells the graph that new messages will always
    # be added to the current states
    messages: Annotated[Sequence[BaseMessage], operator.add]
    # The 'next' field indicates where to route to next
    next: str

class TeamFloAgentStateWithPlan(TypedDict):
    input: str
    plan: List[str]
    past_steps: Annotated[List[Tuple], operator.add]
    response: str


from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from flo_ai.models.flo_executable import ExecutableFlo
from flo_ai.state.flo_state import TeamFloAgentStateWithPlan
from langchain_core.pydantic_v1 import BaseModel, Field
from typing import Union
from langgraph.graph.graph import CompiledGraph
from flo_ai.models.flo_team import FloTeam
from langgraph.graph import StateGraph
from flo_ai.state.flo_session import FloSession
from flo_ai.helpers.utils import randomize_name


class Response(BaseModel):
    """Response to user."""
    response: str

class Plan(BaseModel):
    """Plan to follow in future"""

    steps: list[str] = Field(
        description="different steps to follow, should be in sorted order"
    )

class Act(BaseModel):
    """Action to perform."""
    action: Union[Response, Plan] = Field(
        description="Action to perform. If you want to respond to user, use Response. "
        "If you need to further use tools to get the answer, use Plan."
    )

class FloPlanner(ExecutableFlo):
    def __init__(self, name: str, graph: CompiledGraph) -> None:
        super().__init__(name, graph)

    def draw(self, xray=True):
        return self.graph.get_graph(xray=xray).draw_mermaid_png()
    
    def stream(self, work, config = None):
        return self.graph.stream({
             "input": work
        }, config)
    
    def invoke(self, work):
        return self.graph.invoke({
             "input": work
        })

class FloPlannerBuilder:

    def __init__(self, 
                 session: FloSession,
                 name: str,
                 team: FloTeam,
                 planner_prompt: Union[ChatPromptTemplate, None] = None,
                 replanner_prompt: Union[ChatPromptTemplate, None] = None,
                 llm: Union[BaseLanguageModel, None] = None) -> None:
        self.planner_name = randomize_name(name)
        self.re_planner_name = randomize_name("{}-{}".format("replan", name))
        self.session = session
        self.llm = llm if llm is not None else session.llm
        self.team = team
        planner_prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    """For the given objective, come up with a simple step by step plan. \
        This plan should involve individual tasks, that if executed correctly will yield the correct answer. Do not add any superfluous steps. \
        The result of the final step should be the final answer. Make sure that each step has all the information needed - do not skip steps.""",
                ),
                MessagesPlaceholder(variable_name="messages"),
            ]
        ) if planner_prompt is None else planner_prompt
        self.planner =  planner_prompt | self.llm.with_structured_output(Plan)

        self.replanner_prompt = ChatPromptTemplate.from_template(
                """For the given objective, come up with a simple step by step plan. \
            This plan should involve individual tasks, that if executed correctly will yield the correct answer. Do not add any superfluous steps. \
            The result of the final step should be the final answer. Make sure that each step has all the information needed - do not skip steps.

            Your objective was this:
            {input}

            Your original plan was this:
            {plan}

            You have currently done the follow steps:
            {past_steps}

            Update your plan accordingly. If no more steps are needed and you can return to the user, then do that. 
            Otherwise if there are still steps to complete, fill out the plan. Only add steps to the plan that still NEED to be done. 
            Do not return previously done steps as part of the plan."""
        ) if replanner_prompt is None else replanner_prompt
        self.replanner =  self.replanner_prompt | self.llm.with_structured_output(Act)

    def execute_step(self, state: TeamFloAgentStateWithPlan):
        plan = state["plan"]
        plan_str = "\n".join(f"{i+1}. {step}" for i, step in enumerate(plan))
        task = plan[0]
        task_formatted = f"""For the following plan:
            {plan_str}\n\nYou are tasked with executing step {1}, {task}."""
        agent_response = self.team.invoke(task_formatted)
        return {
            "past_steps": (task, agent_response["messages"][-1].content),
        }

    def plan_step(self, state: TeamFloAgentStateWithPlan):
        plan =  self.planner.invoke({"messages": [("user", state["input"])]})
        if self.session is not None:
            self.session.append(self.planner_name)
        return {"plan": plan.steps}
    
    def replan_step(self, state: TeamFloAgentStateWithPlan):
        output = self.replanner.invoke(state)
        if self.session is not None:
            self.session.append(self.re_planner_name)
        if isinstance(output.action, Response):
            return { "response": output.action.response }
        else:
            return { "plan": output.action.steps }
        
    def should_end(self, state: TeamFloAgentStateWithPlan):
        if "response" in state and state["response"]:
            return "__end__"
        else:
            return self.team.name

    def build(self):
        workflow = StateGraph(TeamFloAgentStateWithPlan)

        # Add the plan node
        workflow.add_node(self.planner_name, self.plan_step)

        # Add the execution step
        workflow.add_node(self.team.name, self.execute_step)

        # Add a replan node
        workflow.add_node(self.re_planner_name, self.replan_step)

        workflow.set_entry_point(self.planner_name)

        # From plan we go to agent
        workflow.add_edge(self.planner_name, self.team.name)

        # From agent, we replan
        workflow.add_edge(self.team.name, self.re_planner_name)

        workflow.add_conditional_edges(
            self.re_planner_name,
            # Next, we pass in the function that will determine which node is called next.
            self.should_end,
        )

        # Finally, we compile it!
        # This compiles it into a LangChain Runnable,
        # meaning you can use it as you would any other runnable
        return FloPlanner(self.planner_name, workflow.compile())


from flo_ai.yaml.flo_team_builder import to_supervised_team
from flo_ai.builders.yaml_builder import build_supervised_team, FloRoutedTeamConfig
from typing import (
    Any,
    Iterator,
    Union
)
from flo_ai.state.flo_session import FloSession
from flo_ai.models.flo_executable import ExecutableFlo

class Flo:

    def __init__(self,
                 session: FloSession,
                 config: FloRoutedTeamConfig) -> None:
        self.config = config
        self.runnable: ExecutableFlo = build_supervised_team(session, config)

    def stream(self, query, config = None) -> Iterator[Union[dict[str, Any], Any]]:
        return self.runnable.stream(query, config)
    
    def invoke(self, query, config = None) -> Iterator[Union[dict[str, Any], Any]]:
        return self.runnable.invoke(query, config)
    
    @staticmethod
    def build(session: FloSession, yaml: str):
        return Flo(session, to_supervised_team(yaml))

    def draw(self, xray=True):
        return self.runnable.draw(xray)
    
    def draw_to_file(self, filename: str, xray=True):
        from PIL import Image as PILImage
        import io
        byte_image = self.runnable.draw(xray)
        with io.BytesIO(byte_image) as image_io:
            image = PILImage.open(image_io)
            image.save(filename)

        


import functools
from langchain_core.messages import HumanMessage
from langgraph.graph.graph import CompiledGraph
from langgraph.graph import StateGraph, END
from langchain.agents import AgentExecutor

from flo_ai.models.flo_executable import ExecutableFlo
from flo_ai.models.flo_agent import FloAgent
from flo_ai.models.flo_supervisor import FloSupervisor
from flo_ai.constants.prompt_constants import FLO_FINISH
from flo_ai.state.flo_state import TeamFloAgentState
from flo_ai.state.flo_session import FloSession
from flo_ai.helpers.utils import randomize_name

class FloAgentNode:
    def __init__(self, 
                 agent_node: functools.partial, 
                 name: str) -> None:
        self.agent_node = agent_node
        self.name = name

def teamflo_agent_node(state: TeamFloAgentState, agent: AgentExecutor, name: str):
    result = agent.invoke(state)
    return {"messages": [HumanMessage(content=result["output"], name=name)]}

# The following functions interoperate between the top level graph state
# and the state of the research sub-graph
# this makes it so that the states of each graph don't get intermixed
def enter_chain(message: str, members: list[str]):
    results = {
        "messages": [HumanMessage(content=message)],
        "team_members": ", ".join(members),
    }
    return results

class FloTeam(ExecutableFlo):
    def __init__(self, name: str, graph: CompiledGraph) -> None:
        super().__init__(name, graph)

    def draw(self, xray=True):
        return self.graph.get_graph(xray=xray).draw_mermaid_png()

class FloTeamChain():
    def __init__(self, name: str, chain) -> None:
        self.chain = chain
        self.name = name

class FloTeamBuilder:
    def __init__(self, 
                 session: FloSession,
                 name: str,
                 supervisor: FloSupervisor) -> None:
        self.name = randomize_name(name)
        self.session = session
        self.flo_agents = supervisor.agents
        self.flo_supervisor: FloSupervisor = supervisor
        self.members = list(map(lambda x: x.name, self.flo_agents))

    def build_node(self, flo_agent: FloAgent):
        agent_func = functools.partial(teamflo_agent_node, agent=flo_agent.executor, name=flo_agent.name)
        return FloAgentNode(agent_func, flo_agent.name)
    

    def build(self):
        if self.flo_supervisor.is_agent_supervisor():
            return self.build_agent_supervisor_graph()
        else:
            return self.build_team_supervisor_graph()
        
    def router(self, state: TeamFloAgentState):
        next = state["next"]
        conditional_map = {k: k for k in self.members}
        conditional_map[FLO_FINISH] = END
        self.session.append(node=next)
        if self.session.is_looping(node=next):
            return conditional_map[FLO_FINISH]
        return conditional_map[next]


    def build_agent_supervisor_graph(self) -> FloTeam:
        flo_agent_nodes = [self.build_node(flo_agent) for flo_agent in self.flo_agents]
        workflow = StateGraph(TeamFloAgentState)
        for flo_agent_node in flo_agent_nodes:
            workflow.add_node(flo_agent_node.name, flo_agent_node.agent_node)
        workflow.add_node(self.flo_supervisor.name, self.flo_supervisor.executor)
        for member in self.members:
            workflow.add_edge(member, self.flo_supervisor.name)

        workflow.add_conditional_edges(self.flo_supervisor.name, self.router)

        workflow.set_entry_point(self.flo_supervisor.name)
        workflow_graph = workflow.compile()
        return FloTeam(self.name, workflow_graph)
    
    def build_chain_for_teams(self, flo_team: FloTeam):
        # TODO lets see if we can convert to members
        return FloTeamChain(flo_team.name, (
            functools.partial(enter_chain, members=flo_team.graph.nodes)
            | flo_team.graph
        ))
    
    def get_last_message(self, state: TeamFloAgentState, second = None) -> str:
        return state["messages"][-1].content


    def join_graph(self, response: dict):
        return {"messages": [response["messages"][-1]]}
    
    def build_team_supervisor_graph(self) -> FloTeam:
        flo_team_entry_chains = [self.build_chain_for_teams(flo_agent) for flo_agent in self.flo_agents]
        # Define the graph.
        super_graph = StateGraph(TeamFloAgentState)
        # First add the nodes, which will do the work
        for flo_team_chain in flo_team_entry_chains:
            super_graph.add_node(flo_team_chain.name, self.get_last_message | flo_team_chain.chain | self.join_graph)
        super_graph.add_node(self.flo_supervisor.name, self.flo_supervisor.executor)

        for member in self.members:
            super_graph.add_edge(member, self.flo_supervisor.name)

        super_graph.add_conditional_edges(self.flo_supervisor.name, self.router)

        super_graph.set_entry_point(self.flo_supervisor.name)
        super_graph = super_graph.compile()
        return FloTeam(self.name, super_graph)




from flo_ai.models.flo_member import FloMember
from langgraph.graph.graph import CompiledGraph
from langchain_core.messages import HumanMessage

class ExecutableFlo(FloMember):
    def __init__(self, name: str, graph: CompiledGraph, type: str = "team") -> None:
        super().__init__(name, type)
        self.graph = graph

    def stream(self, work, config = None):
        return self.graph.stream({
             "messages": [
                HumanMessage(content=work)
            ]
        }, config)
    
    def invoke(self, work, config = None):
        return self.graph.invoke({
             "messages": [
                HumanMessage(content=work)
            ]
        }, config)


from langchain_core.vectorstores import VectorStoreRetriever
from langchain_core.runnables import RunnableParallel, Runnable
from flo_ai.state.flo_session import FloSession
from langchain.schema.output_parser import StrOutputParser
from langchain.schema.runnable import RunnablePassthrough
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from flo_ai.retrievers.flo_multi_query import FloMultiQueryRetriverBuilder
from langchain.retrievers import ContextualCompressionRetriever
from langchain.retrievers.document_compressors import DocumentCompressorPipeline
from flo_ai.retrievers.flo_compression_pipeline import FloCompressionPipeline
from langchain.tools.retriever import create_retriever_tool
from flo_ai.tools.flo_rag_tool import create_flo_rag_tool
from langchain_core.tools import Tool

class FloRagBuilder():
    def __init__(self, session: FloSession, retriever: VectorStoreRetriever) -> None:
        self.session = session
        self.retriever = retriever
        self.default_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", """You are an assistant for question-answering tasks. 
                 Use the following pieces of retrieved context to answer the question. 
                 If you don't know the answer, just say that you don't know. 
                 Use three sentences maximum and keep the answer concise.

                 Here is the context:
                 {context}
                 
                 """),
                MessagesPlaceholder(variable_name="chat_history"),
                ("human", "{question}"),
            ]
        )
        self.history_aware_retriever = self.__create_history_aware_retriever()

    def with_prompt(self, prompt: ChatPromptTemplate):
        self.default_prompt = prompt
        return self

    def with_multi_query(self, prompt = None):
        builder = FloMultiQueryRetriverBuilder(session=self.session,
                                                retriver=self.retriever,
                                                  query_prompt=prompt)
        multi_query_retriever = builder.build()
        self.retriever = multi_query_retriever.retriever
        return self
    
    def with_compression(self, pipeline: FloCompressionPipeline):
        pipeline_compressor = DocumentCompressorPipeline(
            transformers=pipeline.get()
        )
        compression_retriever = ContextualCompressionRetriever(
            base_compressor=pipeline_compressor, base_retriever=self.retriever
        )
        self.retriever = compression_retriever
        return self

    def __create_history_aware_retriever(self):
        contextualize_q_system_prompt = """Given a chat history and the latest user question \
        which might reference context in the chat history, formulate a standalone question \
        which can be understood without the chat history. Do NOT answer the question, \
        just reformulate it if needed and otherwise return it as is."""

        contextualize_q_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", contextualize_q_system_prompt),
                MessagesPlaceholder("chat_history"),
                ("human", "{question}"),
            ]
        )
        self.history_aware_retriever = contextualize_q_prompt | self.session.llm | StrOutputParser()
        return self.history_aware_retriever

    def __get_retriever(self):
        def __precontext_retriver(input_prompt: dict):
            if input_prompt.get("chat_history"):
                return self.history_aware_retriever
            else:
                return input_prompt["question"]
        return __precontext_retriver | self.retriever
    
    def __format_docs(self, docs):
        return "\n\n".join(doc.page_content for doc in docs)
    
    def __get_optional_chat_history(self, x):
        return x["chat_history"] if "chat_history" in x else []
    
    def __build_history_aware_rag(self):
        rag_chain = (
            RunnablePassthrough.assign(
                context=(lambda x: x["context"]),
            )
            | self.default_prompt
            | self.session.llm   
        )

        rag_chain_with_source = RunnableParallel(
            {
                "context": self.__get_retriever() | self.__format_docs,
                "question": RunnablePassthrough(),
                "chat_history": lambda x: self.__get_optional_chat_history(x)
            }
        ).assign(answer=rag_chain)
        return rag_chain_with_source

    def build_rag(self):
        return self.__build_history_aware_rag()
    
    def build_retriever_tool(self, name, description):
        return create_retriever_tool(self.retriever, name, description)
    
    def build_rag_tool(self, name, description) -> Tool:
        rag = self.__build_history_aware_rag()
        return create_flo_rag_tool(rag, name, description)



from langchain_core.tools import BaseTool
from langchain.agents import AgentExecutor
from langchain.agents import create_tool_calling_agent
from langchain_core.runnables import Runnable
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from flo_ai.helpers.utils import randomize_name
from flo_ai.models.flo_executable import ExecutableFlo
from flo_ai.state.flo_session import FloSession
from typing import Union, Optional

class FloAgent(ExecutableFlo):
    def __init__(self, 
                 agent: Runnable, 
                 executor: AgentExecutor, 
                 name: str) -> None:
        super().__init__(name, executor, "agent")
        self.agent: Runnable =  agent,
        self.executor: AgentExecutor = executor

class FloAgentBuilder:
    def __init__(self, 
                 session: FloSession,
                 name: str, 
                 prompt: Union[ChatPromptTemplate, str], 
                 tools: list[BaseTool],
                 verbose: bool = True,
                 role: Optional[str] = None,
                 llm: Union[BaseLanguageModel, None] =  None,
                 return_intermediate_steps: bool = False,
                 handle_parsing_errors: bool = True) -> None:
        self.name: str = randomize_name(name)
        self.llm = llm if llm is not None else session.llm
        # TODO improve to add more context of what other agents are available
        system_prompts = [("system", "You are a {}".format(role)), ("system", prompt)] if role is not None else [("system", prompt)]
        system_prompts.append(MessagesPlaceholder(variable_name="messages"))
        system_prompts.append(MessagesPlaceholder(variable_name="agent_scratchpad"))
        self.prompt: ChatPromptTemplate = ChatPromptTemplate.from_messages(
            system_prompts
        ) if isinstance(prompt, str) else prompt
        self.tools: list[BaseTool] = tools
        self.verbose = verbose
        self.return_intermediate_steps = return_intermediate_steps
        self.handle_parsing_errors = handle_parsing_errors


    def build(self) -> AgentExecutor:
        agent = create_tool_calling_agent(self.llm, self.tools, self.prompt)
        executor = AgentExecutor(agent=agent, 
                             tools=self.tools, 
                             verbose=self.verbose, 
                             return_intermediate_steps=self.return_intermediate_steps, 
                             handle_parsing_errors=self.handle_parsing_errors)
        return FloAgent(agent, executor, self.name)



from pydantic import BaseModel
from typing import List
import yaml
import re
from typing import Optional
from flo_ai.models.exception import FloValidationException


KIND_SUPERVISED_TEAM = "FloRoutedTeam"
KIND_FLO_AGENT = "FloAgent"

yaml_kinds = [
  KIND_SUPERVISED_TEAM,
  KIND_FLO_AGENT
]

class KeyValueArgs(BaseModel):
    name: str
    value: str

class FilterArgs(BaseModel):
    name: str
    description: str
    type: str

class ToolConfig(BaseModel):
    name: str
    args: Optional[List[KeyValueArgs]] = None
    properties: Optional[List[KeyValueArgs]] = None
    filters: Optional[List[FilterArgs]] = None

class AgentConfig(BaseModel):
    name: str
    role: Optional[str] = None
    kind: Optional[str] = None
    job: str
    tools: List[ToolConfig] = []

class RouterConfig(BaseModel):
    name: str
    kind: str

class PlannerConfig(BaseModel):
    name: str

class TeamConfig(BaseModel):
    name: str
    agents: Optional[List[AgentConfig]] = None
    subteams: Optional[List['TeamConfig']] = None
    router: Optional[RouterConfig] = None
    planner: Optional[PlannerConfig] = None

class FloRoutedTeamConfig(BaseModel):
    apiVersion: str
    kind: str
    name: str
    team: TeamConfig

class FloAgentConfig(BaseModel):
    apiVersion: str
    kind: str
    name: str
    agent: AgentConfig

def to_supervised_team(yaml_str: str) -> FloRoutedTeamConfig:
    parsed_data = yaml.safe_load(yaml_str)
    kind = parsed_data["kind"]
    if kind == KIND_SUPERVISED_TEAM:
        flo_supervised_team = FloRoutedTeamConfig(**parsed_data)
        validate_sup_team_config(flo_supervised_team)
        return flo_supervised_team
    elif kind == KIND_FLO_AGENT:
        flo_agent = FloAgentConfig(**parsed_data)
        validate_sup_team_config(flo_agent)
        return flo_agent
    else:
        raise FloValidationException("Unknown kind: {}".format(kind))

def validate_sup_team_config(flo: FloRoutedTeamConfig):
    if flo.kind == KIND_FLO_AGENT:
        return
    if flo.name is None or not is_valid_name(flo.name):
        raise FloValidationException("Invalid agent name while creating the flow, expected: [^[a-z][a-z0-9_-]*$]")
    
def is_valid_name(s: str) -> bool:
    pattern = r'^[a-z][a-z0-9_-]*$'
    return bool(re.match(pattern, s))


from langchain_core.output_parsers.openai_functions import JsonOutputFunctionsParser
from langchain.chains import LLMChain
from langchain_core.language_models import BaseLanguageModel
from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from typing import Union
from flo_ai.models.flo_member import FloMember
from flo_ai.state.flo_session import FloSession
from flo_ai.constants.prompt_constants import FLO_FINISH
from flo_ai.helpers.utils import randomize_name


# TODO, maybe add description about what team members can do
supervisor_system_message = (
    "You are a supervisor tasked with managing a conversation between the"
    " following {member_type}: {members}. Given the following user request,"
    " respond with the worker to act next. Each worker will perform a"
    " task and respond with their results and status. When the users question is answered or the assigned task is finished,"
    " respond with FINISH. If both your workers has responded atleast once, then strictly return FINISH "
)

class StateUpdateComponent:
        def __init__(self, name: str, session: FloSession) -> None:
            self.name = name
            self.inner_session = session

        def __call__(self, input):
            self.inner_session.append(self.name)
            return input


class FloSupervisor:
    def __init__(self,
                 executor: LLMChain, 
                 agents: list[FloMember],
                 name: str) -> None:
        self.executor = executor
        self.agents = agents
        self.name = name
        self.type = agents[0].type

    def is_agent_supervisor(self):
        return self.type == "agent"

class FloSupervisorBuilder:
    def __init__(self,
                 session: FloSession,
                 name: str,
                 reportees: list[FloMember],
                 supervisor_prompt: Union[ChatPromptTemplate, None] = None,
                 llm: Union[BaseLanguageModel, None] = None) -> None:
        # TODO add validation for reporteess
        self.name = randomize_name(name)
        self.session = session
        self.llm = llm if llm is not None else session.llm
        self.agents = reportees
        self.members = [agent.name for agent in reportees]
        self.options = self.members + [FLO_FINISH]
        member_type = "workers" if reportees[0].type == "agent" else "team members"
        self.supervisor_prompt = ChatPromptTemplate.from_messages(
            [
                ("system", supervisor_system_message),
                MessagesPlaceholder(variable_name="messages"),
                (
                    "system",
                    "Given the conversation above, who should act next?"
                    " Or should we FINISH if the task is already answered, (do not wait for the entire question to be answered) or there isn't enough information to answer the question, even then return FINISH? Select one of: {options}",
                ),
            ]
        ).partial(options=str(self.options), members=", ".join(self.members), member_type=member_type) if supervisor_prompt is None else  supervisor_prompt
    
    def build(self):
        function_def = {
            "name": "route",
            "description": "Select the next role.",
            "parameters": {
                "title": "routeSchema",
                "type": "object",
                "properties": {
                    "next": {
                        "title": "Next",
                        "anyOf": [
                            {"enum": self.options},
                        ],
                    }
                },
                "required": ["next"],
            }
        }
            
        chain = (
            self.supervisor_prompt
            | self.llm.bind_functions(functions=[function_def], function_call="route")
            | JsonOutputFunctionsParser()
            | StateUpdateComponent(self.name, self.session)
        )

        return FloSupervisor(chain, self.agents, self.name)


from langchain_core.language_models import BaseLanguageModel
from langchain_core.tools import BaseTool

class FloSession:
    def __init__(self, llm: BaseLanguageModel, loop_size: int = 2, max_loop: int = 3) -> None:
        self.llm = llm
        self.tools = dict()
        self.counter = dict()
        self.navigation: list[str] = list()
        self.pattern_series = dict()
        self.loop_size: int = loop_size
        self.max_loop: int = max_loop

    def register_tool(self, name: str, tool: BaseTool):
        self.tools[name] = tool
        return self

    def append(self, node: str) -> int:
        self.counter[node] = self.counter.get(node, 0) + 1
        if node in self.navigation:
            last_known_index = len(self.navigation) - 1 - self.navigation[::-1].index(node)
            pattern_array = self.navigation[last_known_index: len(self.navigation)]
            if len(pattern_array) + 1 >= self.loop_size:
                pattern = "|".join(pattern_array) + "|" + node
                if node in self.pattern_series:
                    self.pattern_series[node].append(pattern)
                else:
                    self.pattern_series[node] = [pattern]
        self.navigation.append(node)

    def is_looping(self, node) -> bool:
        patterns = self.pattern_series[node] if node in self.pattern_series else []
        if len(patterns) < self.max_loop:
            return False
        return patterns[-(self.max_loop):] == [patterns[-1]] * self.max_loop

    def stringify(self):
        return str(self.counter)


from flo_ai.core import Flo
from flo_ai.models.flo_agent import FloAgentBuilder
from flo_ai.models.flo_supervisor import FloSupervisorBuilder
from flo_ai.models.flo_team import FloTeamBuilder
from flo_ai.state.flo_session import FloSession
from flo_ai.retrievers.flo_retriever import FloRagBuilder



from langchain.pydantic_v1 import BaseModel, Field
from langchain.tools import Tool
from langchain_core.runnables import Runnable
from functools import partial

class FloRagToolInput(BaseModel):
    query: str = Field(description="query to look up in retriever")
    
def __get_rag_answer(query: str, runnable: Runnable):
    result = runnable.invoke({ "question": query })
    return result["answer"].content

async def __aget_rag_answer(query: str, runnable: Runnable):
    result = await runnable.ainvoke({ "question": query })
    return result["answer"].content

def create_flo_rag_tool(
    runnable_rag: Runnable,
    name: str,
    description: str
) -> Tool:
    func = partial(
        __get_rag_answer,
        runnable=runnable_rag
    )

    afunc = partial(
        __aget_rag_answer,
        runnable=runnable_rag
    )
    
    return Tool(
        name=name,
        description=description,
        func=func,
        coroutine=afunc,
        args_schema=FloRagToolInput,
    )

